@@ -13,24 +13,23 @@
 #include "stdafx.h"
 #include "Loaders.h"
 #include "SampleIO.h"
-#include "../soundbase/SampleFormatConverters.h"
-#include "../soundbase/SampleFormatCopy.h"
+#include "openmpt/soundbase/SampleDecode.hpp"
+#include "SampleCopy.h"
+#include "SampleNormalize.h"
 #include "ModSampleCopy.h"
 #include "ITCompression.h"
-#include "../common/mptIO.h"
 #ifndef MODPLUG_NO_FILESAVE
 #include "../common/mptFileIO.h"
+#include "mpt/io/base.hpp"
+#include "mpt/io/io.hpp"
+#include "mpt/io/io_stdstream.hpp"
+#include "mpt/io_write/buffer.hpp"
 #endif
-#include <stdexcept>
+#include "BitReader.h"
 
 
 OPENMPT_NAMESPACE_BEGIN
 
-// Sample decompression routines in other source files
-void AMSUnpack(mpt::const_byte_span source, mpt::byte_span dest, int8 packCharacter);
-uint8 MDLReadBits(uint32 &bitbuf, int32 &bitnum, const uint8 *(&ibuf), size_t &bytesLeft, int8 n);
-uintptr_t DMFUnpack(uint8 *psample, const uint8 *ibuf, const uint8 *ibufmax, uint32 maxlen);
-
 
 // Read a sample from memory
 size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
@@ -45,8 +44,8 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 	FileReader::off_t bytesRead = 0;	// Amount of memory that has been read from file
 
 	FileReader::off_t filePosition = file.GetPosition();
-	const mpt::byte * sourceBuf = nullptr;
-	FileReader::PinnedRawDataView restrictedSampleDataView;
+	const std::byte * sourceBuf = nullptr;
+	FileReader::PinnedView restrictedSampleDataView;
 	FileReader::off_t fileSize = 0;
 	if(UsesFileReaderForDecoding())
 	{
@@ -54,22 +53,16 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 		fileSize = file.BytesLeft();
 	} else if(!IsVariableLengthEncoded())
 	{
-		restrictedSampleDataView = file.GetPinnedRawDataView(CalculateEncodedSize(sample.nLength));
+		restrictedSampleDataView = file.GetPinnedView(CalculateEncodedSize(sample.nLength));
 		sourceBuf = restrictedSampleDataView.data();
 		fileSize = restrictedSampleDataView.size();
 		if(sourceBuf == nullptr)
 			return 0;
 	} else
 	{
-		// Only DMF sample compression encoding should fall in this case,
-		MPT_ASSERT(GetEncoding() == DMF);
-		// file is guaranteed by the caller to be ONLY data for this sample,
-		// it is thus efficient to create a view to the whole file object.
-		// See MPT_ASSERT with fileSize below.
-		restrictedSampleDataView = file.GetPinnedRawDataView();
-		sourceBuf = restrictedSampleDataView.data();
-		fileSize = restrictedSampleDataView.size();
+		MPT_ASSERT_NOTREACHED();
 	}
+
 	if(!IsVariableLengthEncoded() && sample.nLength > 0x40000)
 	{
 		// Limit sample length to available bytes in file to avoid excessive memory allocation.
@@ -96,6 +89,19 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 			maxLength /= encodedBytesPerSample;
 		}
 		LimitMax(sample.nLength, mpt::saturate_cast<SmpLength>(maxLength));
+	} else if(GetEncoding() == IT214 || GetEncoding() == IT215 || GetEncoding() == MDL || GetEncoding() == DMF)
+	{
+		// In the best case, IT compression represents each sample point as a single bit.
+		// In practice, there is of course the two-byte header per compressed block and the initial bit width change.
+		// As a result, if we have a file length of n, we know that the sample can be at most n*8 sample points long.
+		// For DMF, there are at least two bits per sample, and for MDL at least 5 (so both are worse than IT).
+		size_t maxLength = fileSize;
+		uint8 maxSamplesPerByte = 8 / GetNumChannels();
+		if(Util::MaxValueOfType(maxLength) / maxSamplesPerByte >= maxLength)
+			maxLength *= maxSamplesPerByte;
+		else
+			maxLength = Util::MaxValueOfType(maxLength);
+		LimitMax(sample.nLength, mpt::saturate_cast<SmpLength>(maxLength));
 	} else if(GetEncoding() == AMS)
 	{
 		if(fileSize <= 9)
@@ -110,7 +116,7 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 			maxLength *= 85;
 		else
 			maxLength = Util::MaxValueOfType(maxLength);
-		LimitMax(sample.nLength, maxLength / (GetBitDepth() / 8u));
+		LimitMax(sample.nLength, maxLength / (m_bitdepth / 8u));
 	}
 
 	if(sample.nLength < 1)
@@ -143,7 +149,7 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 			LimitMax(readLength, file.BytesLeft());
 
 			const uint8 *inBuf = mpt::byte_cast<const uint8*>(sourceBuf) + sizeof(compressionTable);
-			int8 *outBuf = sample.pSample8;
+			int8 *outBuf = sample.sample8();
 			int8 delta = 0;
 
 			for(size_t i = readLength; i != 0; i--)
@@ -169,19 +175,17 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 		int8 packCharacter = file.ReadUint8();
 		bytesRead += 9;
 
-		FileReader::PinnedRawDataView packedDataView = file.ReadPinnedRawDataView(sourceSize);
+		FileReader::PinnedView packedDataView = file.ReadPinnedView(sourceSize);
 		LimitMax(sourceSize, mpt::saturate_cast<uint32>(packedDataView.size()));
 		bytesRead += sourceSize;
 
-		AMSUnpack(packedDataView.span(), mpt::as_span(mpt::void_cast<mpt::byte*>(sample.pSample), sample.GetSampleSizeInBytes()), packCharacter);
+		AMSUnpack(packedDataView.span(), mpt::as_span(sample.sampleb(), sample.GetSampleSizeInBytes()), packCharacter);
 		if(sample.uFlags[CHN_16BIT] && !mpt::endian_is_little())
 		{
-			int16 *p = sample.pSample16;
+			auto p = sample.sample16();
 			for(SmpLength length = sample.nLength; length != 0; length--, p++)
 			{
-				int16le v;
-				std::memcpy(&v, p, sizeof(int16));
-				*p = v;
+				*p = mpt::bit_cast<int16le>(*p);
 			}
 		}
 	} else if(GetEncoding() == PTM8Dto16 && GetChannelFormat() == mono && GetBitDepth() == 16)
@@ -193,16 +197,8 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 		// Huffman MDL compressed samples
 		if(file.CanRead(8) && (fileSize = file.ReadUint32LE()) >= 4)
 		{
-			FileReader chunk = file.ReadChunk(fileSize);
+			BitReader chunk = file.ReadChunk(fileSize);
 			bytesRead = chunk.GetLength() + 4;
-			uint32 bitBuf = chunk.ReadUint32LE();
-			int32 bitNum = 32;
-
-			restrictedSampleDataView = chunk.GetPinnedRawDataView();
-			sourceBuf = restrictedSampleDataView.data();
-
-			const uint8 *inBuf = reinterpret_cast<const uint8*>(sourceBuf);
-			size_t bytesLeft = chunk.BytesLeft();
 
 			uint8 dlt = 0, lowbyte = 0;
 			const bool is16bit = GetBitDepth() == 16;
@@ -213,20 +209,20 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 					uint8 hibyte;
 					if(is16bit)
 					{
-						lowbyte = MDLReadBits(bitBuf, bitNum, inBuf, bytesLeft, 8);
+						lowbyte = static_cast<uint8>(chunk.ReadBits(8));
 					}
-					bool sign = MDLReadBits(bitBuf, bitNum, inBuf, bytesLeft, 1) != 0;
-					if(MDLReadBits(bitBuf, bitNum, inBuf, bytesLeft, 1))
+					bool sign = chunk.ReadBits(1) != 0;
+					if(chunk.ReadBits(1))
 					{
-						hibyte = MDLReadBits(bitBuf, bitNum, inBuf, bytesLeft, 3);
+						hibyte = static_cast<uint8>(chunk.ReadBits(3));
 					} else
 					{
 						hibyte = 8;
-						while(!MDLReadBits(bitBuf, bitNum, inBuf, bytesLeft, 1))
+						while(!chunk.ReadBits(1))
 						{
 							hibyte += 0x10;
 						}
-						hibyte += MDLReadBits(bitBuf, bitNum, inBuf, bytesLeft, 4);
+						hibyte += static_cast<uint8>(chunk.ReadBits(4));
 					}
 					if(sign)
 					{
@@ -235,14 +231,13 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 					dlt += hibyte;
 					if(!is16bit)
 					{
-						sample.pSample8[j] = dlt;
-					}
-					else
+						sample.sample8()[j] = dlt;
+					} else
 					{
-						sample.pSample16[j] = lowbyte | (dlt << 8);
+						sample.sample16()[j] = lowbyte | (dlt << 8);
 					}
 				}
-			} catch(const std::range_error &)
+			} catch(const BitReader::eof &)
 			{
 				// Data is not sufficient to decode the whole sample
 				//AddToLog(LogWarning, "Truncated MDL sample block");
@@ -253,108 +248,33 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 		// DMF Huffman compression
 		if(fileSize > 4)
 		{
-			const uint8 *inBuf = mpt::byte_cast<const uint8*>(sourceBuf);
-			const uint8 *inBufMax = inBuf + fileSize;
-			uint8 *outBuf = static_cast<uint8 *>(sample.pSample);
-			bytesRead = DMFUnpack(outBuf, inBuf, inBufMax, sample.GetSampleSizeInBytes());
-
-			// This assertion ensures that, when using variable length samples,
-			// the caller actually provided a trimmed chunk to read the sample data from.
-			// This is required as we cannot know the encoded sample data size upfront
-			// to construct a properly sized pinned view.
-			MPT_ASSERT(bytesRead == fileSize);
-
+			bytesRead = DMFUnpack(file, mpt::byte_cast<uint8 *>(sample.sampleb()), sample.GetSampleSizeInBytes());
 		}
-#ifdef MODPLUG_TRACKER
 	} else if((GetEncoding() == uLaw || GetEncoding() == aLaw) && GetBitDepth() == 16 && (GetChannelFormat() == mono || GetChannelFormat() == stereoInterleaved))
 	{
-		// 8-to-16 bit G.711 u-law / a-law
-		static const int16 uLawTable[256] =
-		{
-			-32124,-31100,-30076,-29052,-28028,-27004,-25980,-24956,
-			-23932,-22908,-21884,-20860,-19836,-18812,-17788,-16764,
-			-15996,-15484,-14972,-14460,-13948,-13436,-12924,-12412,
-			-11900,-11388,-10876,-10364, -9852, -9340, -8828, -8316,
-			 -7932, -7676, -7420, -7164, -6908, -6652, -6396, -6140,
-			 -5884, -5628, -5372, -5116, -4860, -4604, -4348, -4092,
-			 -3900, -3772, -3644, -3516, -3388, -3260, -3132, -3004,
-			 -2876, -2748, -2620, -2492, -2364, -2236, -2108, -1980,
-			 -1884, -1820, -1756, -1692, -1628, -1564, -1500, -1436,
-			 -1372, -1308, -1244, -1180, -1116, -1052,  -988,  -924,
-			  -876,  -844,  -812,  -780,  -748,  -716,  -684,  -652,
-			  -620,  -588,  -556,  -524,  -492,  -460,  -428,  -396,
-			  -372,  -356,  -340,  -324,  -308,  -292,  -276,  -260,
-			  -244,  -228,  -212,  -196,  -180,  -164,  -148,  -132,
-			  -120,  -112,  -104,   -96,   -88,   -80,   -72,   -64,
-			   -56,   -48,   -40,   -32,   -24,   -16,    -8,    -1,
-			 32124, 31100, 30076, 29052, 28028, 27004, 25980, 24956,
-			 23932, 22908, 21884, 20860, 19836, 18812, 17788, 16764,
-			 15996, 15484, 14972, 14460, 13948, 13436, 12924, 12412,
-			 11900, 11388, 10876, 10364,  9852,  9340,  8828,  8316,
-			  7932,  7676,  7420,  7164,  6908,  6652,  6396,  6140,
-			  5884,  5628,  5372,  5116,  4860,  4604,  4348,  4092,
-			  3900,  3772,  3644,  3516,  3388,  3260,  3132,  3004,
-			  2876,  2748,  2620,  2492,  2364,  2236,  2108,  1980,
-			  1884,  1820,  1756,  1692,  1628,  1564,  1500,  1436,
-			  1372,  1308,  1244,  1180,  1116,  1052,   988,   924,
-			   876,   844,   812,   780,   748,   716,   684,   652,
-			   620,   588,   556,   524,   492,   460,   428,   396,
-			   372,   356,   340,   324,   308,   292,   276,   260,
-			   244,   228,   212,   196,   180,   164,   148,   132,
-			   120,   112,   104,    96,    88,    80,    72,    64,
-			    56,    48,    40,    32,    24,    16,     8,     0,
-		};
-
-		static const int16 aLawTable[256] =
-		{
-			 -5504, -5248, -6016, -5760, -4480, -4224, -4992, -4736,
-			 -7552, -7296, -8064, -7808, -6528, -6272, -7040, -6784,
-			 -2752, -2624, -3008, -2880, -2240, -2112, -2496, -2368,
-			 -3776, -3648, -4032, -3904, -3264, -3136, -3520, -3392,
-			-22016,-20992,-24064,-23040,-17920,-16896,-19968,-18944,
-			-30208,-29184,-32256,-31232,-26112,-25088,-28160,-27136,
-			-11008,-10496,-12032,-11520, -8960, -8448, -9984, -9472,
-			-15104,-14592,-16128,-15616,-13056,-12544,-14080,-13568,
-			  -344,  -328,  -376,  -360,  -280,  -264,  -312,  -296,
-			  -472,  -456,  -504,  -488,  -408,  -392,  -440,  -424,
-			   -88,   -72,   -120,  -104,  -24,   -8,    -56,   -40,
-			  -216,  -200,  -248,  -232,  -152,  -136,  -184,  -168,
-			 -1376, -1312, -1504, -1440, -1120, -1056, -1248, -1184,
-			 -1888, -1824, -2016, -1952, -1632, -1568, -1760, -1696,
-			  -688,  -656,  -752,  -720,  -560,  -528,  -624,  -592,
-			  -944,  -912,  -1008, -976,  -816,  -784,  -880,  -848,
-			  5504,  5248,  6016,  5760,  4480,  4224,  4992,  4736,
-			  7552,  7296,  8064,  7808,  6528,  6272,  7040,  6784,
-			  2752,  2624,  3008,  2880,  2240,  2112,  2496,  2368,
-			  3776,  3648,  4032,  3904,  3264,  3136,  3520,  3392,
-			 22016, 20992, 24064, 23040, 17920, 16896, 19968, 18944,
-			 30208, 29184, 32256, 31232, 26112, 25088, 28160, 27136,
-			 11008, 10496, 12032, 11520,  8960,  8448,  9984,  9472,
-			 15104, 14592, 16128, 15616, 13056, 12544, 14080, 13568,
-			   344,   328,   376,   360,   280,   264,   312,   296,
-			   472,   456,   504,   488,   408,   392,   440,   424,
-			    88,    72,   120,   104,    24,     8,    56,    40,
-			   216,   200,   248,   232,   152,   136,   184,   168,
-			  1376,  1312,  1504,  1440,  1120,  1056,  1248,  1184,
-			  1888,  1824,  2016,  1952,  1632,  1568,  1760,  1696,
-			   688,   656,   752,   720,   560,   528,   624,   592,
-			   944,   912,  1008,   976,   816,   784,   880,   848,
-		};
-
-		const int16 *lut = GetEncoding() == uLaw ? uLawTable : aLawTable;
 
 		SmpLength readLength = sample.nLength * GetNumChannels();
-		LimitMax(readLength, file.BytesLeft());
+		LimitMax(readLength, mpt::saturate_cast<SmpLength>(fileSize));
 		bytesRead = readLength;
 
-		const uint8 *inBuf = mpt::byte_cast<const uint8*>(sourceBuf);
-		int16 *outBuf = sample.pSample16;
+		const std::byte *inBuf = sourceBuf;
+		int16 *outBuf = sample.sample16();
 
-		while(readLength--)
+		if(GetEncoding() == uLaw)
 		{
-			*(outBuf++) = lut[*(inBuf++)];
+			SC::DecodeInt16uLaw conv;
+			while(readLength--)
+			{
+				*(outBuf++) = conv(inBuf++);
+			}
+		} else
+		{
+			SC::DecodeInt16ALaw conv;
+			while(readLength--)
+			{
+				*(outBuf++) = conv(inBuf++);
+			}
 		}
-#endif // MODPLUG_TRACKER
 	}
 
 
@@ -377,9 +297,6 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 		case MT2:
 			bytesRead = CopyMonoSample<SC::DecodeInt8Delta>(sample, sourceBuf, fileSize);
 			break;
-		case PCM7to8:		// 7 Bit stored as 8-Bit with highest bit unused / Mono / Signed / PCM
-			bytesRead = CopyMonoSample<SC::DecodeInt7>(sample, sourceBuf, fileSize);
-			break;
 		default:
 			MPT_ASSERT_NOTREACHED();
 			break;
@@ -399,13 +316,14 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 			bytesRead = CopyStereoSplitSample<SC::DecodeUint8>(sample, sourceBuf, fileSize);
 			break;
 		case deltaPCM:		// 8-Bit / Stereo Split / Delta / PCM
+		case MT2:			// same as deltaPCM, but right channel is stored as a difference from the left channel
 			bytesRead = CopyStereoSplitSample<SC::DecodeInt8Delta>(sample, sourceBuf, fileSize);
-			break;
-		case MT2:		// same as deltaPCM, but right channel is stored as a difference from the left channel
-			bytesRead = CopyStereoSplitSample<SC::DecodeInt8Delta>(sample, sourceBuf, fileSize);
-			for(SmpLength i = 0; i < sample.nLength * 2; i += 2)
+			if(GetEncoding() == MT2)
 			{
-				sample.pSample8[i + 1] = static_cast<int8>(static_cast<uint8>(sample.pSample8[i + 1]) + static_cast<uint8>(sample.pSample8[i]));
+				for(int8 *p = sample.sample8(), *pEnd = p + sample.nLength * 2; p < pEnd; p += 2)
+				{
+					p[1] = static_cast<int8>(static_cast<uint8>(p[0]) + static_cast<uint8>(p[1]));
+				}
 			}
 			break;
 		default:
@@ -491,13 +409,14 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 			bytesRead = CopyStereoSplitSample<SC::DecodeInt16<0x8000u, littleEndian16> >(sample, sourceBuf, fileSize);
 			break;
 		case deltaPCM:		// 16-Bit / Stereo Split / Delta / PCM
-			bytesRead = CopyStereoSplitSample<SC::DecodeInt16Delta<littleEndian16> >(sample, sourceBuf, fileSize);
-			break;
 		case MT2:		// same as deltaPCM, but right channel is stored as a difference from the left channel
 			bytesRead = CopyStereoSplitSample<SC::DecodeInt16Delta<littleEndian16> >(sample, sourceBuf, fileSize);
-			for(SmpLength i = 0; i < sample.nLength * 2; i += 2)
+			if(GetEncoding() == MT2)
 			{
-				sample.pSample16[i + 1] = static_cast<int16>(static_cast<uint16>(sample.pSample16[i + 1]) + static_cast<uint16>(sample.pSample16[i]));
+				for(int16 *p = sample.sample16(), *pEnd = p + sample.nLength * 2; p < pEnd; p += 2)
+				{
+					p[1] = static_cast<int16>(static_cast<uint16>(p[0]) + static_cast<uint16>(p[1]));
+				}
 			}
 			break;
 		default:
@@ -757,7 +676,7 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 		if(bytesRead && srcPeak != 1.0f)
 		{
 			// Adjust sample volume so we do not affect relative volume of the sample. Normalizing is only done to increase precision.
-			sample.nGlobalVol = Util::Round<uint16>(Clamp(sample.nGlobalVol * srcPeak, 1.0f, 64.0f));
+			sample.nGlobalVol = mpt::saturate_round<uint16>(Clamp(sample.nGlobalVol * srcPeak, 1.0f, 64.0f));
 			sample.uFlags.set(SMP_MODIFIED);
 		}
 	}
@@ -778,7 +697,7 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 		if(bytesRead && srcPeak != 1.0)
 		{
 			// Adjust sample volume so we do not affect relative volume of the sample. Normalizing is only done to increase precision.
-			sample.nGlobalVol = Util::Round<uint16>(Clamp(sample.nGlobalVol * srcPeak, 1.0f, 64.0f));
+			sample.nGlobalVol = mpt::saturate_round<uint16>(Clamp(sample.nGlobalVol * srcPeak, 1.0, 64.0));
 			sample.uFlags.set(SMP_MODIFIED);
 		}
 	}
@@ -884,30 +803,36 @@ size_t SampleIO::ReadSample(ModSample &sample, FileReader &file) const
 
 
 // Write a sample to file
-size_t SampleIO::WriteSample(std::ostream *f, const ModSample &sample, SmpLength maxSamples) const
+size_t SampleIO::WriteSample(std::ostream &f, const ModSample &sample, SmpLength maxSamples) const
 {
-	if(!sample.HasSampleData()) return 0;
-
-	union
+	if(sample.uFlags[CHN_ADLIB])
 	{
-		int8 buffer8[8192];
-		int16 buffer16[4096];
-	};
-	const void *const pSampleVoid = sample.pSample;
-	const int8 *const pSample8 = sample.pSample8;
-	const int16 *const pSample16 = sample.pSample16;
+		mpt::IO::Write(f, sample.adlib);
+		return sizeof(sample.adlib);
+	}
+	if(!sample.HasSampleData())
+	{
+		return 0;
+	}
+
+	std::array<std::byte, mpt::IO::BUFFERSIZE_TINY> writeBuffer;
+	mpt::IO::WriteBuffer<std::ostream> fb{f, mpt::as_span(writeBuffer)};
+
 	SmpLength numSamples = sample.nLength;
 
-	if(maxSamples && numSamples > maxSamples) numSamples = maxSamples;
+	if(maxSamples && numSamples > maxSamples)
+	{
+		numSamples = maxSamples;
+	}
 
-	size_t len = CalculateEncodedSize(numSamples), bufcount = 0;
+	std::size_t len = CalculateEncodedSize(numSamples);
 
 	if(GetBitDepth() == 16 && GetChannelFormat() == mono && GetEndianness() == littleEndian &&
 		(GetEncoding() == signedPCM || GetEncoding() == unsignedPCM || GetEncoding() == deltaPCM))
 	{
 		// 16-bit little-endian mono samples
 		MPT_ASSERT(len == numSamples * 2);
-		if(!f) return len;
+		const int16 *const pSample16 = sample.sample16();
 		const int16 *p = pSample16;
 		int s_old = 0;
 		const int s_ofs = (GetEncoding() == unsignedPCM) ? 0x8000 : 0;
@@ -923,20 +848,13 @@ size_t SampleIO::WriteSample(std::ostream *f, const ModSample &sample, SmpLength
 			}
 			if(GetEncoding() == deltaPCM)
 			{
-				buffer16[bufcount] = SwapBytesLE((int16)(s_new - s_old));
+				mpt::IO::Write(fb, mpt::as_le(static_cast<int16>(s_new - s_old)));
 				s_old = s_new;
 			} else
 			{
-				buffer16[bufcount] = SwapBytesLE((int16)(s_new + s_ofs));
-			}
-			bufcount++;
-			if(bufcount >= mpt::size(buffer16))
-			{
-				mpt::IO::WriteRaw(*f, reinterpret_cast<mpt::byte*>(buffer16), bufcount * 2);
-				bufcount = 0;
+				mpt::IO::Write(fb, mpt::as_le(static_cast<int16>(s_new + s_ofs)));
 			}
 		}
-		if (bufcount) mpt::IO::WriteRaw(*f, reinterpret_cast<mpt::byte*>(buffer16), bufcount * 2);
 	}
 
 	else if(GetBitDepth() == 8 && GetChannelFormat() == stereoSplit &&
@@ -944,33 +862,25 @@ size_t SampleIO::WriteSample(std::ostream *f, const ModSample &sample, SmpLength
 	{
 		// 8-bit Stereo samples (not interleaved)
 		MPT_ASSERT(len == numSamples * 2);
-		if(!f) return len;
+		const int8 *const pSample8 = sample.sample8();
 		const int s_ofs = (GetEncoding() == unsignedPCM) ? 0x80 : 0;
 		for (uint32 iCh=0; iCh<2; iCh++)
 		{
 			const int8 *p = pSample8 + iCh;
 			int s_old = 0;
-
-			bufcount = 0;
-			for (uint32 j=0; j<numSamples; j++)
+			for (SmpLength j = 0; j < numSamples; j++)
 			{
 				int s_new = *p;
 				p += 2;
 				if (GetEncoding() == deltaPCM)
 				{
-					buffer8[bufcount++] = (int8)(s_new - s_old);
+					mpt::IO::Write(fb, static_cast<int8>(s_new - s_old));
 					s_old = s_new;
 				} else
 				{
-					buffer8[bufcount++] = (int8)(s_new + s_ofs);
-				}
-				if(bufcount >= mpt::size(buffer8))
-				{
-					mpt::IO::WriteRaw(*f, reinterpret_cast<mpt::byte*>(buffer8), bufcount);
-					bufcount = 0;
+					mpt::IO::Write(fb, static_cast<int8>(s_new + s_ofs));
 				}
 			}
-			if (bufcount) mpt::IO::WriteRaw(*f, reinterpret_cast<mpt::byte*>(buffer8), bufcount);
 		}
 	}
 
@@ -979,65 +889,94 @@ size_t SampleIO::WriteSample(std::ostream *f, const ModSample &sample, SmpLength
 	{
 		// 16-bit little-endian Stereo samples (not interleaved)
 		MPT_ASSERT(len == numSamples * 4);
-		if(!f) return len;
+		const int16 *const pSample16 = sample.sample16();
 		const int s_ofs = (GetEncoding() == unsignedPCM) ? 0x8000 : 0;
 		for (uint32 iCh=0; iCh<2; iCh++)
 		{
 			const int16 *p = pSample16 + iCh;
 			int s_old = 0;
-
-			bufcount = 0;
-			for (SmpLength j=0; j<numSamples; j++)
+			for (SmpLength j = 0; j < numSamples; j++)
 			{
 				int s_new = *p;
 				p += 2;
 				if (GetEncoding() == deltaPCM)
 				{
-					buffer16[bufcount] = SwapBytesLE((int16)(s_new - s_old));
+					mpt::IO::Write(fb, mpt::as_le(static_cast<int16>(s_new - s_old)));
 					s_old = s_new;
 				} else
 				{
-					buffer16[bufcount] = SwapBytesLE((int16)(s_new + s_ofs));
-				}
-				bufcount++;
-				if(bufcount >= mpt::size(buffer16))
-				{
-					mpt::IO::WriteRaw(*f, reinterpret_cast<mpt::byte*>(buffer16), bufcount * 2);
-					bufcount = 0;
+					mpt::IO::Write(fb, mpt::as_le(static_cast<int16>(s_new + s_ofs)));
 				}
 			}
-			if (bufcount) mpt::IO::WriteRaw(*f, reinterpret_cast<mpt::byte*>(buffer16), bufcount * 2);
 		}
 	}
 
-	else if((GetBitDepth() == 8 || (GetBitDepth() == 16 && GetEndianness() == GetNativeEndianness())) && GetChannelFormat() == stereoInterleaved && GetEncoding() == signedPCM)
+	else if(GetBitDepth() == 8 && GetChannelFormat() == stereoInterleaved && GetEncoding() == signedPCM)
 	{
 		// Stereo signed interleaved
-		if(f) mpt::IO::WriteRaw(*f, mpt::void_cast<const mpt::byte*>(pSampleVoid), len);
+		MPT_ASSERT(len == numSamples * 2);
+		const int8 *const pSample8 = sample.sample8();
+		mpt::IO::Write(f, mpt::as_span(pSample8, len));
+	}
+
+	else if(GetBitDepth() == 16 && GetChannelFormat() == stereoInterleaved && GetEncoding() == signedPCM && GetEndianness() == littleEndian)
+	{
+		// Stereo signed interleaved, little-endian
+		MPT_ASSERT(len == numSamples * 4);
+		const int16 *const pSample16 = sample.sample16();
+		const int16 *p = pSample16;
+		for(SmpLength j = 0; j < numSamples; j++)
+		{
+			mpt::IO::Write(fb, mpt::as_le(p[0]));
+			mpt::IO::Write(fb, mpt::as_le(p[1]));
+			p += 2;
+		}
+	}
+
+	else if(GetBitDepth() == 16 && GetChannelFormat() == stereoInterleaved && GetEncoding() == signedPCM && GetEndianness() == bigEndian)
+	{
+		// Stereo signed interleaved, big-endian
+		MPT_ASSERT(len == numSamples * 4);
+		const int16 *const pSample16 = sample.sample16();
+		const int16 *p = pSample16;
+		for(SmpLength j = 0; j < numSamples; j++)
+		{
+			mpt::IO::Write(fb, mpt::as_be(p[0]));
+			mpt::IO::Write(fb, mpt::as_be(p[1]));
+			p += 2;
+		}
+	}
+
+	else if(GetBitDepth() == 16 && GetChannelFormat() == stereoSplit && GetEncoding() == signedPCM && GetEndianness() == bigEndian)
+	{
+		// Stereo signed split, big-endian
+		MPT_ASSERT(len == numSamples * 4);
+		for(uint8 chn = 0; chn < 2; chn++)
+		{
+			const int16 *p = sample.sample16() + chn;
+			for(SmpLength j = 0; j < numSamples; j++)
+			{
+				mpt::IO::Write(fb, mpt::as_be(*p));
+				p += 2;
+			}
+		}
 	}
 
 	else if(GetBitDepth() == 8 && GetChannelFormat() == stereoInterleaved && GetEncoding() == unsignedPCM)
 	{
 		// Stereo unsigned interleaved
 		MPT_ASSERT(len == numSamples * 2);
-		if(!f) return len;
-		for(SmpLength j = 0; j < len; j++)
+		const int8 *const pSample8 = sample.sample8();
+		for(SmpLength j = 0; j < numSamples * 2; j++)
 		{
-			buffer8[bufcount] = (int8)((uint8)(pSample8[j]) + 0x80);
-			bufcount++;
-			if(bufcount >= mpt::size(buffer8))
-			{
-				mpt::IO::WriteRaw(*f, reinterpret_cast<mpt::byte*>(buffer8), bufcount);
-				bufcount = 0;
-			}
+			mpt::IO::Write(fb, static_cast<int8>(static_cast<uint8>(pSample8[j]) + 0x80));
 		}
-		if (bufcount) mpt::IO::WriteRaw(*f, reinterpret_cast<mpt::byte*>(buffer8), bufcount);
 	}
 
 	else if(GetEncoding() == IT214 || GetEncoding() == IT215)
 	{
 		// IT2.14-encoded samples
-		ITCompression its(sample, GetEncoding() == IT215, f, numSamples);
+		ITCompression its(sample, GetEncoding() == IT215, &f, numSamples);
 		len = its.GetCompressedSize();
 	}
 
@@ -1046,60 +985,60 @@ size_t SampleIO::WriteSample(std::ostream *f, const ModSample &sample, SmpLength
 	{
 		MPT_ASSERT(GetBitDepth() == 8);
 		MPT_ASSERT(len == numSamples);
-		if(!f) return len;
-		const int8 *p = pSample8;
-		int sinc = sample.GetElementarySampleSize();
-		int s_old = 0;
-		const int s_ofs = (GetEncoding() == unsignedPCM) ? 0x80 : 0;
-		MPT_MAYBE_CONSTANT_IF(mpt::endian_is_little())
+		if(sample.uFlags[CHN_16BIT])
 		{
-			if (sample.uFlags[CHN_16BIT]) p++;
-		}
-
-		for (SmpLength j = 0; j < len; j++)
+			const int16 *p = sample.sample16();
+			int s_old = 0;
+			const int s_ofs = (GetEncoding() == unsignedPCM) ? 0x80 : 0;
+			for(SmpLength j = 0; j < numSamples; j++)
+			{
+				int s_new = mpt::rshift_signed(*p, 8);
+				p++;
+				if(sample.uFlags[CHN_STEREO])
+				{
+					s_new = (s_new + mpt::rshift_signed(*p, 8) + 1) / 2;
+					p++;
+				}
+				if(GetEncoding() == deltaPCM)
+				{
+					mpt::IO::Write(fb, static_cast<int8>(s_new - s_old));
+					s_old = s_new;
+				} else
+				{
+					mpt::IO::Write(fb, static_cast<int8>(s_new + s_ofs));
+				}
+			}
+		} else
 		{
-			int s_new = (int8)(*p);
-			p += sinc;
-			if (sample.uFlags[CHN_STEREO])
+			const int8 *const pSample8 = sample.sample8();
+			const int8 *p = pSample8;
+			int s_old = 0;
+			const int s_ofs = (GetEncoding() == unsignedPCM) ? 0x80 : 0;
+			for(SmpLength j = 0; j < numSamples; j++)
 			{
-				s_new = (s_new + ((int)*p) + 1) / 2;
-				p += sinc;
-			}
-			if (GetEncoding() == deltaPCM)
-			{
-				buffer8[bufcount++] = (int8)(s_new - s_old);
-				s_old = s_new;
-			} else
-			{
-				buffer8[bufcount++] = (int8)(s_new + s_ofs);
-			}
-			if(bufcount >= mpt::size(buffer8))
-			{
-				mpt::IO::WriteRaw(*f, reinterpret_cast<mpt::byte*>(buffer8), bufcount);
-				bufcount = 0;
+				int s_new = *p;
+				p++;
+				if(sample.uFlags[CHN_STEREO])
+				{
+					s_new = (s_new + (static_cast<int>(*p)) + 1) / 2;
+					p++;
+				}
+				if(GetEncoding() == deltaPCM)
+				{
+					mpt::IO::Write(fb, static_cast<int8>(s_new - s_old));
+					s_old = s_new;
+				} else
+				{
+					mpt::IO::Write(fb, static_cast<int8>(s_new + s_ofs));
+				}
 			}
 		}
-		if (bufcount) mpt::IO::WriteRaw(*f, reinterpret_cast<mpt::byte*>(buffer8), bufcount);
 	}
+
 	return len;
 }
 
 
-// Write a sample to file
-size_t SampleIO::WriteSample(std::ostream &f, const ModSample &sample, SmpLength maxSamples) const
-{
-	return WriteSample(&f, sample, maxSamples);
-}
-
-
-// Write a sample to file
-size_t SampleIO::WriteSample(FILE *f, const ModSample &sample, SmpLength maxSamples) const
-{
-	mpt::FILE_ostream s(f);
-	return WriteSample(f ? &s : nullptr, sample, maxSamples);
-}
-
-
 #endif // MODPLUG_NO_FILESAVE
 
 
